from nexus.tools.check_zeros import check_zeros, remove_zeros
from nexus.tools.combine_nexuses import combine_nexuses
from nexus.tools.shufflenexus import shufflenexus
from nexus.tools.sites import iter_constant_sites
from nexus.tools.sites import iter_unique_sites
from nexus.tools.sites import count_site_values
from nexus.tools.sites import new_nexus_without_sites
from nexus.tools.sites import tally_by_site
from nexus.tools.sites import tally_by_taxon
from nexus.tools.sites import count_binary_set_size
from nexus.tools.trees import (
    delete_trees, sample_trees, strip_comments_in_trees, visit_trees, visit_tree_nodes)

__all__ = [
    "binarise",
    "multistatise",
    "combine_nexuses",
    "shufflenexus",
    "iter_constant_sites",
    "iter_unique_sites",
    "count_site_values",
    "new_nexus_without_sites",
    "tally_by_site",
    "tally_by_taxon",
    "count_binary_set_size",
    "check_zeros",
    "remove_zeros",
    "delete_trees",
    "sample_trees",
    "strip_comments_in_trees",
    "visit_trees",
    "visit_tree_nodes",
]



import re
import typing

from clldutils.text import strip_brackets, split_text_with_context
import newick

from nexus.handlers import GenericHandler
from nexus.exceptions import NexusFormatException, TranslateTableException


class Tree(str):
    @classmethod
    def from_newick(cls,
                    new: typing.Union[str, newick.Node],
                    name: str = 'tree',
                    rooted: typing.Union[bool, None] = None):
        rooting = '' if rooted is None else '[&{}] '.format('R' if rooted else 'U')
        new = new if isinstance(new, str) else new.newick
        return cls('tree {} = {}{}{}'.format(name, rooting, new, '' if new.endswith(';') else ';'))

    @property
    def name(self):
        m = re.search(r'tree\s+(?P<name>[^=]+)\s*=', strip_brackets(self, {'[': ']'}))
        if m:
            return m.group('name').strip()

    @property
    def rooted(self):
        """
        Three-valued attribute, specifying whether tree is rooted, unrooted or unspecified.

        :return: `True`|`False`|`None`
        """
        m = re.search(r'\[&(?P<rooting>[RU])]', self)
        if m:
            return m.group('rooting') == 'R'
        return None

    @property
    def newick_string(self):
        # Find the string up to the first "(" which is not inside a comment ...
        prefix = split_text_with_context(self, separators='(', brackets={'[': ']'})[0]
        # ... the remainder of the line should be the newick representation of the tree:
        return self[len(prefix):].strip()

    @property
    def newick_tree(self):
        return newick.loads(self.newick_string)[0]


class TreeHandler(GenericHandler):
    """Handler for `trees` blocks"""
    is_tree = re.compile(r"""tree\s+.*=.*;""", re.IGNORECASE)

    translate_regex = re.compile(r"""
        ([,(])              # boundary
        ([A-Z0-9_\-\.]+)    # taxa-id
        :?                  # optional colon
        (\[.+?\])?          # minimally match an optional comment chunk
        (\d+(\.\d+)?)?      # optional branchlengths
        (?=[),])?           # end boundary
    """, re.IGNORECASE + re.VERBOSE + re.DOTALL)

    translate_regex_beast = re.compile(r"""
        ([,(])              # boundary
        ([A-Z0-9_\-\.]+)    # taxa-id
        (\[.+?\])?          # minimally match an optional comment chunk
        :?                  # optional colon
        (\[.+?\])           # minimally match an optional comment chunk
        :?                  # optional colon
        (\d+(\.\d+)?)       # optional branchlengths
        (?=[),])?           # end boundary
    """, re.IGNORECASE + re.VERBOSE + re.DOTALL)

    def __init__(self, **kw):
        super(TreeHandler, self).__init__(**kw)
        # does the treefile have a translate block?
        self.was_translated = False
        # has detranslate been called?
        self._been_detranslated = False
        self.translators = {}
        self.attributes = []
        self.trees = []

        translate_start = re.compile(r"""^translate$""", re.IGNORECASE)
        translation_pattern = re.compile(r"""(\d+)\s(['"\w\d\*\.\_\-]+)[,;]?""")

        lost_in_translation = False
        for line in self.block:
            # look for translation start, and turn on lost_in_translation
            if translate_start.match(line):
                lost_in_translation = True
                self.was_translated = True
            elif self.is_mesquite_attribute(line):
                self.attributes.append(line)

            # if we're in a translate block
            elif lost_in_translation:
                if translation_pattern.match(line):
                    taxon_id, taxon = translation_pattern.findall(line)[0]
                    taxon = taxon.strip("'")
                    if taxon_id in self.translators:
                        raise NexusFormatException(
                            "Duplicate Taxa ID %s in translate block" % taxon_id
                        )
                    if taxon in self.translators.values():
                        raise NexusFormatException(
                            "Duplicate Taxon %s in translate block" % taxon
                        )
                    self.translators[taxon_id] = taxon
                if line.endswith(';'):
                    lost_in_translation = False

            elif self.is_tree.search(line):
                self.trees.append(Tree(line))

        # if there is not translate block then get the list of taxa from the first tree.
        if (not self.translators) and self.trees:
            # remove comments first to avoid issues with comments like
            # "var(median) = y"
            onetree = self.remove_comments(self.trees[0])
            taxa = re.findall(r"""[(),](\w+)[:),]""", onetree)
            for taxon_id, t in enumerate(taxa, 1):
                self.translators[taxon_id] = t

    def __getitem__(self, index):
        return self.trees[index]

    @property
    def taxa(self):
        return self.translators.values()

    @property
    def ntrees(self):
        return len(self.trees)

    def detranslate(self):
        """Detranslates all trees in the file"""
        if self._been_detranslated:
            return
        for idx, tree in enumerate(self.trees):
            self.trees[idx] = Tree(self._detranslate_tree(tree, self.translators))
        self._been_detranslated = True

    @staticmethod
    def _findall_chunks(tree):
        """Helper function to find groups used by detranslate."""

        # check for beast tree or not -> decide which regex to use
        if '{' in tree and '[' in tree:  # sufficient?
            regex = TreeHandler.translate_regex_beast
        else:
            regex = TreeHandler.translate_regex

        matches = []
        index = 0
        while True:
            match = regex.search(tree, index)
            if not match:
                break
            m = dict(zip(['start', 'taxon', 'comment', 'branch'], match.groups()))
            m['match'] = tree[match.start():match.end() + 1]
            m['end'] = tree[match.end()]
            matches.append(m)
            index = match.end()
        return matches

    def _detranslate_tree(self, tree, translatetable):
        """
        Takes a `tree` and expands the short format tree with translated
        taxa labels from `translatetable` into a full format tree.

        :param tree: String containing newick tree
        :type tree: String

        :param translatetable: Mapping of taxa id -> taxa names
        :type translatetable: Dict

        :return: String of detranslated tree
        """
        i = 0
        for i, found in enumerate(self._findall_chunks(tree), start=1):
            if found['taxon'] in translatetable:
                taxon = translatetable[found['taxon']]
                if found['comment'] and found['branch']:
                    # comment and branch
                    sub = "%s:%s%s" % \
                        (taxon, found['comment'], found['branch'])
                elif found['comment']:
                    # comment only
                    sub = "%s%s" % (taxon, found['comment'])
                elif found['branch']:
                    # branch only
                    sub = "%s:%s" % (taxon, found['branch'])
                else:
                    # taxon only
                    sub = taxon
                sub = "%s%s%s" % (found['start'], sub, found['end'])
                tree = tree.replace(found['match'], sub)
        if len(translatetable) and len(translatetable) != i:
            raise TranslateTableException(
                "Mismatch between translate table size (n={}) and expected taxa in trees "
                "(n={})".format(len(translatetable), i))
        return tree

    def iter_lines(self):
        for attr in self.attributes:
            yield "\t" + attr
        if self.was_translated and not self._been_detranslated:
            yield '\ttranslate'
            translator_keys = [int(k) for k in self.translators.keys()]
            for i, index in enumerate(sorted(translator_keys), start=1):
                yield "\t%d %s%s" % (
                    index, self.translators[str(index)], '' if i == len(translator_keys) else ',')
            # work around bug https://github.com/CompEvol/beast2/issues/713
            yield ';'
        for tree in self.trees:
            yield "\t" + tree



"""
Tools for writing a nexus file
"""
import collections

from nexus.util import FileWriterMixin

TEMPLATE = """\
#NEXUS
%(datablock)s
%(treeblock)s
"""

DATA_TEMPLATE = """
%(comments)s

BEGIN DATA;
  DIMENSIONS NTAX=%(ntax)d NCHAR=%(nchar)d;
  FORMAT DATATYPE=%(datatype)s MISSING=%(missing)s GAP=%(gap)s %(interleave)s SYMBOLS="%(symbols)s";
  %(charblock)s
MATRIX
%(collabels)s%(matrix)s%(collabels)s
;
END;
"""

TREE_TEMPLATE = """
BEGIN TREES;
%(trees)s
END;
"""


class NexusWriter(FileWriterMixin):

    MISSING = '?'
    GAP = '-'
    DATATYPE = 'STANDARD'

    def __init__(self):
        self.comments = []
        self.collabels = []
        self._taxa = None
        self._characters = None
        self.data = collections.defaultdict(dict)
        self.is_binary = False
        self.trees = []
        self._taxa_in = []
        self._chars_in = []
        self.preserve_order = False
        self.padding = 3

    def clean(self, s):
        """Removes unsafe characters"""
        _EMPTY = ''
        replacements = {
            ' ': _EMPTY, '\\': _EMPTY, ':': _EMPTY,
            '/': _EMPTY, '?': _EMPTY, '-': _EMPTY,
            '(': '_', ')': _EMPTY,
        }
        for f, t in replacements.items():
            s = s.replace(f, t)
        return s

    @property
    def characters(self):
        if self._characters is None:
            self._characters = list(sorted(self.data.keys(), key=lambda x: self._chars_in.index(x)
                                           if self.preserve_order else x))
        return self._characters

    @property
    def ntrees(self):
        return len(self.trees)

    @property
    def taxa(self):
        if self._taxa is None:
            self._taxa = set()
            [self._taxa.update(self.data[c].keys()) for c in self.data]
            self._taxa = list(sorted(self._taxa, key=lambda x: self._taxa_in.index(x)
                                     if self.preserve_order else x))
        return self._taxa

    @property
    def symbols(self):
        symbols = set()
        [symbols.update(self.data[c].values()) for c in self.data]
        symbols = [s for s in symbols if s not in ('-', '?')]
        return symbols

    def _iter_charlabels(self):
        """Generates a character label block"""
        chars_len = len(self.characters)
        yield "CHARSTATELABELS"
        for i, char in enumerate(self.characters, 1):
            yield "    %d %s%s" % (
                i, self.clean(str(char)), '' if i == chars_len else ',')
        yield ";"

    def _iter_matrix(self, interleave):
        """Generates a matrix block"""
        max_taxon_size = max([len(t) for t in self.taxa]) + self.padding

        if interleave:
            for c in self.characters:
                for t in self.taxa:
                    yield "%s %s" % (t.ljust(max_taxon_size), self.data[c].get(t, self.MISSING))
                yield ""
        else:
            for t in self.taxa:
                s = []
                for c in self.characters:
                    value = self.data[c].get(t, self.MISSING)
                    if len(value) > 1:  # wrap equivocal states in ()'s
                        value = "(%s)" % value
                    s.append(value)
                yield "%s %s" % (t.ljust(max_taxon_size), ''.join(s))

    def make_treeblock(self):
        return "\n".join(["    %s" % t.lstrip().strip() for t in self.trees])

    def _make_comments(self):
        """Generates a comments block"""
        return "\n".join(["[%s]" % c.ljust(70) if len(c) else "" for c in self.comments])

    def add_comment(self, comment):
        """Adds a `comment` into the nexus file"""
        self.comments.append(comment)

    def _make_collabels(self):
        """Generates a matrix column labels block as comment"""
        pad = " " * (max([len(t) for t in self.taxa]) + self.padding)
        return "\n".join(["%s[%s]" % (pad, c) if len(c) else "" for c in self.collabels])

    def add_collabels(self, collabel):
        """
        Adds a `matrix column label` as comment into the nexus file
        (each row has to be calculated by the user in beforehand,
        padding will be done automatically)
        """
        if isinstance(collabel, list):
            self.collabels.extend(collabel)
        else:
            self.collabels.append(collabel)

    def add(self, taxon, character, value, check=False):
        """Adds a `character` for the given `taxon` and sets it to `value`"""
        assert self.is_binary is False, "Unable to add data to a binarised nexus form"
        if check:
            assert isinstance(character, (str, int)), 'Character must not be of type {}'.format(
                type(character))
            if self.characters:
                assert all(isinstance(c, str) for c in [character] + self.characters) or \
                    all(isinstance(c, int) for c in [character] + self.characters), \
                    "Characters of mixed type are not supported"
        value = str(value)

        if taxon not in self._taxa_in:
            self._taxa_in.append(taxon)
        if character not in self.data:
            self._chars_in.append(character)

        self._taxa = None
        self._characters = None

        # have multiple entries
        if taxon in self.data[character]:
            self.data[character][taxon] += value
        else:
            self.data[character][taxon] = value

    def remove(self, taxon, character):
        """Removes a `character` for the given `taxon` and sets it to empty"""
        del(self.data[character][taxon])
        self._taxa = None
        self._characters = None
        for char in self.data:
            if taxon in self.data[char]:
                break
        else:
            self._taxa_in.remove(taxon)

    def remove_taxon(self, taxon):
        """Removes a given `taxon` from the nexus file"""
        for char in self.data:
            del(self.data[char][taxon])
        self._taxa = None
        self._taxa_in.remove(taxon)

    def remove_character(self, character):
        """Removes a given `character` from the nexus file"""
        del(self.data[character])
        self._characters = None
        self._chars_in.remove(character)

    def write(self, interleave=False, charblock=False, preserve_order=False, **kw):
        """
        Generates a string representation of the nexus
        (basically a wrapper around make_nexus)

        :param interleave: Generate interleaved matrix or not
        :type interleave: Boolean
        :param charblock: Include a characters block or not
        :type charblock: Boolean
        :param preserve_order: Preserve input order of taxa and characters or not
        :type preserve_order: Boolean

        :return: String
        """
        return self.make_nexus(interleave, charblock, preserve_order)

    def _is_valid(self):
        """Checks the nexus is valid to write (i.e. not empty)"""
        if self.data and self.taxa:
            return True
        if self.ntrees:
            return True
        return False

    def make_nexus(self, interleave=False, charblock=False, preserve_order=False):
        """
        Generates a string representation of the nexus

        :param interleave: Generate interleaved matrix or not
        :type interleave: Boolean
        :param charblock: Include a characters block or not
        :type charblock: Boolean
        :param preserve_order: Preserve input order of taxa and characters or not
        :type preserve_order: Boolean

        :return: String
        """
        self.preserve_order = preserve_order

        if not self._is_valid():
            raise ValueError("Nexus has no data!")

        if self.data:
            datablock = DATA_TEMPLATE % {
                'ntax': len(self.taxa),
                'nchar': len(self.characters),
                'charblock': '\n'.join(self._iter_charlabels()) if charblock else '',
                'matrix': '\n'.join(self._iter_matrix(interleave=interleave)),
                'interleave': 'INTERLEAVE' if interleave else '',
                'comments': self._make_comments(),
                'symbols': ''.join(sorted(self.symbols)),
                'collabels': self._make_collabels() if self.collabels else '',
                'missing': self.MISSING,
                'gap': self.GAP,
                'datatype': self.DATATYPE,
            }
        else:
            datablock = ""

        treeblock = TREE_TEMPLATE % {'trees': self.make_treeblock()} if self.ntrees else ""
        return TEMPLATE % {'datablock': datablock, 'treeblock': treeblock}

    def write_as_table(self, preserve_order=False):
        """
        Generates a simple table of the nexus

        :param preserve_order: Preserve input order of taxa and characters or not
        :type preserve_order: Boolean
        """
        self.preserve_order = preserve_order
        out = []
        for t in self.taxa:
            s = []
            for c in self.characters:
                value = self.data[c].get(t, self.MISSING)
                if len(value) > 1:  # wrap equivocal states in ()'s
                    value = "(%s)" % value
                s.append(value)
            out.append("%s %s" % (t.ljust(25), ''.join(s)))
        return "\n".join(out)

    def _convert_to_reader(self):
        """
        This is a hack to convert a NexusWriter object to a NexusReader
        instance.

        One day I'll refactor this all so Reader and Writer subclass something,
        which will make this unnecessary.
        """
        from .reader import NexusReader
        return NexusReader.from_string(
            self.make_nexus(interleave=False, charblock=True, preserve_order=False))



"""
Tools for reading a nexus file
"""
import io
import gzip
import pathlib
import warnings

from nexus.handlers import GenericHandler
from nexus.handlers import BEGIN_PATTERN, END_PATTERN
from nexus.handlers.taxa import TaxaHandler
from nexus.handlers.data import CharacterHandler, DataHandler
from nexus.handlers.tree import TreeHandler
from nexus.exceptions import NexusFormatException

HANDLERS = {
    'data': DataHandler,
    'characters': CharacterHandler,
    'trees': TreeHandler,
    'taxa': TaxaHandler,
}


class NexusReader(object):
    """A nexus reader"""
    def __init__(self, filename=None, **blocks):
        assert not (filename and blocks), 'cannot initialize from file *and* blocks'
        self.filename = None
        self.short_filename = None
        self.blocks = {}
        self.data = None
        self.characters = None
        self.trees = None
        self.taxa = None

        if blocks:
            self._set_blocks(blocks)

        if filename:
            self.filename = filename
            self.short_filename = pathlib.Path(filename).name
            self._set_blocks(NexusReader._blocks_from_file(filename))

    @classmethod
    def from_file(cls, filename, encoding='utf-8-sig'):
        """
        Loads and Parses a Nexus File

        :param filename: filename of a nexus file
        :raises IOError: If file reading fails.
        :return: `NexusReader` object.
        """
        res = cls()
        res._set_blocks(NexusReader._blocks_from_file(filename, encoding=encoding))
        res.filename = filename
        res.short_filename = pathlib.Path(filename).name
        return res

    @classmethod
    def from_string(cls, string):
        """
        Loads and Parses a Nexus from a string

        :param contents: string or string-like object containing a nexus
        :type contents: string

        :return: None
        """
        res = cls()
        res._set_blocks(NexusReader._blocks_from_string(string))
        return res

    def _set_blocks(self, blocks):
        self.blocks = {}
        for block, lines in (blocks.items() if isinstance(blocks, dict) else blocks):
            if block in self.blocks:
                raise NexusFormatException("Duplicate Block %s" % block)
            self.blocks[block] = HANDLERS.get(block, GenericHandler)(name=block, data=lines)

        if self.blocks.get('characters') and not self.blocks.get('data'):
            self.blocks['data'] = self.blocks['characters']

        # copy character labels to data block if data block doesn't have labels
        if self.blocks.get('characters') and self.blocks.get('data'):
            if self.blocks['characters'].charlabels and not self.blocks.get('data').charlabels:
                self.blocks['data'].charlabels = self.blocks['characters'].charlabels

        for block in self.blocks:
            setattr(self, block, self.blocks[block])

    def read_file(self, filename, encoding='utf-8-sig'):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated method NexusReader.read_file",
            category=DeprecationWarning,
            stacklevel=1)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        self._set_blocks(NexusReader._blocks_from_file(filename, encoding=encoding))
        self.filename = filename
        self.short_filename = pathlib.Path(filename).name

    def read_string(self, contents):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated method NexusReader.read_string",
            category=DeprecationWarning,
            stacklevel=1)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        self._set_blocks(NexusReader._blocks_from_string(contents))
        return self

    @staticmethod
    def _blocks_from_string(string):
        return NexusReader._iter_blocks(io.StringIO(string).readlines())

    @staticmethod
    def _iter_blocks(iterlines):
        block, lines = None, []

        for line in iterlines:
            line = line.strip()
            if not line:
                continue
            elif line.startswith('[') and line.endswith(']'):
                continue

            start = BEGIN_PATTERN.findall(line)
            if start:
                if block and lines:
                    # "end" is optional!
                    yield block, lines
                block, lines = start[0][0].lower(), []

            if block:
                lines.append(line)

            if END_PATTERN.search(line):
                if block:
                    yield block, lines
                block, lines = None, []

        if block and lines:
            # "end" is optional. Whatever we have left is counted as belonging to the last block.
            yield block, lines

    @staticmethod
    def _blocks_from_file(filename, encoding='utf-8-sig'):
        filename = pathlib.Path(filename)

        if not (filename.exists() and filename.is_file()):
            raise IOError("Unable To Read File %s" % filename)

        if filename.suffix == '.gz':
            handle = gzip.open(str(filename), 'rt', encoding=encoding)
        else:
            handle = filename.open('r', encoding=encoding)
        res = NexusReader._iter_blocks(handle.readlines())
        handle.close()
        return res

    def write(self, **kw):
        """
        Generates a string containing a complete nexus from
        all the data.

        :return: String
        """
        out = ["#NEXUS\n"]
        blocks = []
        for block in self.blocks.values():
            if block in blocks:
                # We skip copies of blocks - which might have happened to provide shortcuts for
                # semantically identical but differently named blocks, such as "data" and
                # "characters".
                continue
            blocks.append(block)
        for block in blocks:
            out.append(block.write())
            # empty line after block if needed
            if len(blocks) > 1:
                out.append("\n")
        return "\n".join(out)

    def write_to_file(self, filename):
        """
        Writes the nexus to a file.

        :return: None
        """
        with pathlib.Path(filename).open('w', encoding='utf8') as handle:
            handle.writelines(self.write())



from nexus.writer import NexusWriter
from .util import with_nexus_reader


@with_nexus_reader
def multistatise(nexus_obj, charlabel=None):
    """
    Returns a multistate variant of the given `nexus_obj`.

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A NexusReader instance
    :raises AssertionError: if nexus_obj is not a nexus
    :raises NexusFormatException: if nexus_obj does not have a `data` block
    """
    charlabel = charlabel or getattr(nexus_obj, 'short_filename', 1) or 1

    states = {}
    for taxon in nexus_obj.data.matrix:
        states[taxon] = []
        sequence = nexus_obj.data.matrix[taxon]
        for site_idx, value in enumerate(sequence):
            if site_idx > 26:
                raise ValueError("Too many characters to handle! - run out of A-Z")
            assert value == str(value), "%r is not a string" % value
            if value == '1':
                states[taxon].append(chr(65 + site_idx))

    nexout = NexusWriter()
    for taxon in states:
        if not states[taxon]:
            nexout.add(taxon, charlabel, '?')
        else:
            for s in states[taxon]:
                nexout.add(taxon, charlabel, s)
    return nexout._convert_to_reader()



import os

from nexus import NexusWriter
from .util import get_nexus_reader


def combine_nexuses(*nexuslist):
    """
    Combines a list of NexusReader instances into a single nexus

    :param nexuslist: A list of NexusReader instances

    :return: A NexusWriter instance
    """
    if len(nexuslist) == 1 and isinstance(nexuslist[0], (list, tuple)):
        nexuslist = nexuslist[0]
    nexuslist = [get_nexus_reader(nex) for nex in nexuslist]
    out = NexusWriter()
    # check they're all nexus instances and get all block types
    blocks = set()
    for nex in nexuslist:
        blocks.update(list(nex.blocks))

    for block in blocks:
        if block == 'data':
            out = combine_datablocks(out, nexuslist)
        elif block == 'trees':
            out = combine_treeblocks(out, nexuslist)
        else:  # pragma: no cover
            raise ValueError("Don't know how to combine %s blocks" % block)
    return out


def combine_treeblocks(out, nexuslist):
    for nex in nexuslist:
        out.trees.extend(nex.trees.trees)
    return out


def combine_datablocks(out, nexuslist):
    charpos = 0
    for nex_id, nex in enumerate(nexuslist, 1):
        if nex.short_filename:
            nexus_label = os.path.splitext(nex.short_filename)[0]
        else:
            nexus_label = str(nex_id)

        out.add_comment("%d - %d: %s" % (charpos, charpos + nex.data.nchar - 1, nexus_label))
        # handle data
        for site_idx, site in enumerate(sorted(nex.data.characters), 0):
            data = nex.data.characters.get(site)
            charpos += 1
            # work out character label
            charlabel = nex.data.charlabels.get(site_idx, site_idx + 1)
            label = '%s.%s' % (nexus_label, charlabel)

            for taxon, value in data.items():
                out.add(taxon, label, value)
    return out



"""Contains Nexus Manipulation Tools that operate on Site/Characters"""
import collections

from nexus.writer import NexusWriter
from .util import with_nexus_reader


@with_nexus_reader
def iter_constant_sites(nexus_obj):
    """
    Returns a list of zero-based indices of the constant sites in a nexus
    """
    for i in range(0, nexus_obj.data.nchar):
        if len({data[i] for _, data in nexus_obj.data if data[i] not in {'?', '-'}}) == 1:
            yield i


@with_nexus_reader
def iter_unique_sites(nexus_obj):
    """
    Returns a list of the unique sites in a binary nexus
    i.e. sites with only one taxon belonging to them.
        (this only really makes sense if the data is coded as presence/absence)
    """
    for i in range(0, nexus_obj.data.nchar):
        members = collections.Counter()
        missing = 0
        for taxa, characters in nexus_obj.data:
            c = characters[i]
            if c in ('?', '-'):
                missing += 1
            else:
                members.update([c])

        # a character is unique if there's only two states
        # AND there's a state with 1 member
        # AND the state with 1 member is NOT the 0 (absence) state
        if len(members) == 2:
            for state, count in members.items():
                if state != '0' and count == 1:
                    yield i


def count_site_values(nexus_obj, characters=('-', '?')):
    """
    Counts the number of sites with values in `characters` in a nexus

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param characters: An iterable of the characters to count
    :type characters: tuple

    :return: (A dictionary of taxa and missing counts, and a log)
    :raises AssertionError: if nexus_obj is not a nexus
    :raises NexusFormatException: if nexus_obj does not have a `data` block
    """
    tally = {taxon: 0 for taxon, _ in nexus_obj.data}
    for taxon, sites in nexus_obj.data:
        for site in sites:
            if site in characters:
                tally[taxon] += 1
    return tally


def new_nexus_without_sites(nexus_obj, sites_to_remove):
    """
    Returns a new NexusReader instance with the sites in
    `sites_to_remove` removed.

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param sites_to_remove: A list of site numbers
    :type sites_to_remove: List

    :return: A NexusWriter instance
    :raises AssertionError: if nexus_obj is not a nexus
    :raises NexusFormatException: if nexus_obj does not have a `data` block
    """
    # make new nexus
    nexout = NexusWriter()
    nexout.add_comment(
        "Removed %d sites: %s" %
        (len(sites_to_remove), ",".join(["%s" % s for s in sites_to_remove]))
    )
    new_sitepos = 0
    for sitepos in range(nexus_obj.data.nchar):
        if sitepos in sites_to_remove:
            continue  # skip!
        for taxon, data in nexus_obj.data:
            nexout.add(taxon, new_sitepos, data[sitepos])
        new_sitepos += 1
    return nexout


def tally_by_site(nexus_obj):
    """
    Counts the number of taxa per state per site (i.e. site 1 has three taxa
    coded as "A", and 1 taxa coded as "G")

    Returns a dictionary of the cognate sets by members in the nexus

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A Dictionary
    :raises AssertionError: if nexus_obj is not a nexus
    :raises NexusFormatException: if nexus_obj does not have a `data` block

    e.g. {
        'site1': {'state1': ['taxon1', 'taxon2'], 'state0': ['taxon3'], }
        'site2': {'state1': ['taxon2'], 'state0': ['taxon1', 'taxon3'], }
    }
    """
    tally = {}
    for site, data in nexus_obj.data.characters.items():
        tally[site] = tally.get(site, {})
        for taxon, state in data.items():
            tally[site][state] = tally[site].get(state, [])
            tally[site][state].append(taxon)
    return tally


def tally_by_taxon(nexus_obj):
    """
    Counts the number of states per site that each taxon has (i.e. taxon 1
    has three sites coded as "A" and 1 coded as "G")

    Returns a dictionary of the cognate sets by members in the nexus

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A Dictionary
    :raises AssertionError: if nexus_obj is not a nexus
    :raises NexusFormatException: if nexus_obj does not have a `data` block

    e.g. {
        'taxon1': {'state1': ['site1', 'site2'], 'state0': ['site3'], }
        'taxon2': {'state1': ['site2'], 'state0': ['site1', 'site3'], }
    }
    """
    tally = {}
    for taxon, characters in nexus_obj.data:
        tally[taxon] = {}
        for pos, char in enumerate(characters):
            label = nexus_obj.data.charlabels.get(pos, pos)
            tally[taxon][char] = tally[taxon].get(char, [])
            tally[taxon][char].append(label)
    return tally


def count_binary_set_size(nexus_obj):
    """
    Counts the number of sites by their size (i.e. how many sites have two
    members, etc)

    Returns a dictionary of the set size and count

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A Dictionary

    e.g. {
        0: 0,
        1: 100,
        2: 20,
    }
    """
    tally = collections.Counter()
    for char_id in nexus_obj.data.characters:
        char = nexus_obj.data.characters[char_id]
        tally[len([v for v in char.values() if v == '1'])] += 1
    return tally



"""
Converts binary nexuses to a multistate nexus.
"""
import argparse

from nexus.tools import multistatise, combine_nexuses
from nexus.cli_util import add_nexus, get_reader, add_output, write_output


def register(parser):
    parser.add_argument('--charblock', help=argparse.SUPPRESS, action='store_false', default=True)
    parser.add_argument('--interleave', help=argparse.SUPPRESS, action='store_true', default=False)
    add_output(parser)
    add_nexus(parser, many=True)


def run(args):
    nexuslist = [multistatise.multistatise(n) for n in get_reader(args, many=True)]
    write_output(combine_nexuses(nexuslist), args)



"""
combines a series of nexuses into one nexus.
"""
from nexus.tools import combine_nexuses
from nexus.cli_util import add_nexus, get_reader, add_output, write_output


def register(parser):
    add_output(parser)
    add_nexus(parser, many=True)


def run(args):
    write_output(combine_nexuses(get_reader(args, many=True)), args)



"""
Shuffles the characters between each taxon to create a new nexus
"""
from nexus.tools import shufflenexus
from nexus.cli_util import add_nexus, get_reader, add_output, write_output


def register(parser):
    add_nexus(parser)
    add_output(parser)
    parser.add_argument(
        "-n", "--numchars",
        default=0,
        type=int,
        help="Number of Characters to Generate")


def run(args):
    write_output(shufflenexus(get_reader(args), args.numchars or False), args)



import string
import statistics
import collections

SAFE_CHARACTERS = string.ascii_letters + string.digits + '-_'


class Checker(object):

    EMPTY_STATES = ('?', '-', '0')

    def __init__(self, nex):
        self.errors, self.messages = [], []
        self.check(nex)

    @property
    def has_errors(self):
        return len(self.errors) > 0

    def check(self, nex):  # pragma: no cover
        raise NotImplementedError("Should be subclassed")

    def log(self, message):
        self.messages.append(message)

    def status(self):  # pragma: no cover
        print("%s\t%d errors" % (self.__class__.__name__.ljust(50), len(self.errors)))
        for i, e in enumerate(self.messages, 1):
            print("\t%3d. %s" % (i, e))
        for i, e in enumerate(self.errors, 1):
            print("\t%3d. %s" % (i, e))


class DuplicateLabelChecker(Checker):
    """
    Checks for Duplicate Character Labels
    """
    def check(self, nex):
        labels = collections.Counter(nex.data.charlabels.values())
        for label in labels:
            if labels[label] > 1:
                self.errors.append("Duplicate Label (n=%d): %s" % (labels[label], label))
        return not self.has_errors


class PotentiallyUnsafeTaxaLabelsChecker(Checker):
    """
    Checks for Potentially Unsafe Taxa Labels (non-Latin 1 characters etc)
    """
    def check(self, nex):
        for taxon in nex.data.taxa:
            bad = [char for char in taxon if char not in SAFE_CHARACTERS]
            if len(bad):
                self.errors.append(
                    "Potentially Unsafe Taxon Label: %s: %s" % (taxon, " ".join(bad)))
        return not self.has_errors


class LabelChecker(Checker):
    """
    Checks that the number of character labels (if present) is correct
    """
    def check(self, nex):
        nlabels = len(nex.data.charlabels)
        if nlabels == 0:
            return not self.has_errors
        if nlabels != nex.data.nchar:
            self.errors.append(
                "Incorrect number of character labels (expected %d, got %d)" % (
                    nex.data.nchar, nlabels))
        return not self.has_errors


class UnusualStateChecker(Checker):
    """
    Checks for unusual states.

    Returns errors if there are states with counts of less than `THRESHOLD`.
    """
    THRESHOLD = 0.001  # anything less than this is flagged

    def check(self, nex):
        states = collections.Counter()
        for s in nex.data.matrix.values():
            states.update(s)
        total = sum(states.values())
        for s, n in states.most_common():
            if n <= total * self.THRESHOLD:
                self.errors.append("Unusual State Found: %s (n=%d)" % (s, n))
        return not self.has_errors


class EmptyCharacterChecker(Checker):
    """
    Checks for Empty Characters (i.e. sites with nothing other than the values in
    `EMPTY_STATES`. By default: ('?', '-', '0')
    """
    MIN_COUNT = 0

    def check(self, nex):
        tally = collections.Counter()
        for taxon in nex.data.matrix:
            tally.update([
                i for i, c in enumerate(nex.data.matrix[taxon], 0)
                if c not in self.EMPTY_STATES
            ])

        for i in range(0, nex.data.nchar):
            n = tally.get(i, 0)
            if n == self.MIN_COUNT:
                if 'ascert' in nex.data.charlabels.get(i, '').lower():  # ignore
                    self.log("Character %d is an ascertainment character" % i)
                else:
                    self.errors.append("Character %d has count %d" % (i, self.MIN_COUNT))
        return not self.has_errors


class SingletonCharacterChecker(EmptyCharacterChecker):
    MIN_COUNT = 1


class LowStateCountChecker(Checker):
    """
    Checks for taxa with low character states.

    Returns errors if there are taxa with counts of less than 3 standard
    deviations
    """
    THRESHOLD = 3  # 3 x the standard deviation

    def check(self, nex):
        counts = {}
        for taxon in nex.data.matrix:
            counts[taxon] = len([
                c for c in nex.data.matrix[taxon] if c not in self.EMPTY_STATES])

        med = statistics.median(counts.values())
        sd = statistics.stdev(counts.values())
        sd_threshold = med - (self.THRESHOLD * sd)

        for taxon in sorted(counts):
            if counts[taxon] <= sd_threshold:
                self.errors.append(
                    "Taxon %s has a low state count (%d, median = %0.2f - %0.2f)" % (
                        taxon, counts[taxon], med, sd_threshold))
        return not self.has_errors


class BEASTAscertainmentChecker(Checker):
    """
    Checks that the ascertainment correction for BEAST is correct
    """
    def check(self, nex):
        # do we have character labels? -- find characters identified as _asc
        ascert = [c for c in nex.data.charlabels if 'ascert' in nex.data.charlabels[c]]
        # are they empty?
        for a in ascert:
            states = [nex.data.matrix[t][a] for t in nex.data.matrix]
            states = collections.Counter([s for s in states if s not in self.EMPTY_STATES])
            if len(states):
                self.errors.append(
                    "Character %d - %s should be an ascertainment character but has data (%r)" % (
                        a, nex.data.charlabels[a], states))

        # if we have more than one then we should have assumptions block
        if len(ascert) > 1:
            if 'assumptions' not in [k.lower() for k in nex.blocks.keys()]:
                self.errors.append("Missing assumptions block")
        return not self.has_errors


# TODO check assumptions block

CHECKERS = {
    'base': [
        LabelChecker,
        DuplicateLabelChecker,
        PotentiallyUnsafeTaxaLabelsChecker,
        LowStateCountChecker,
        EmptyCharacterChecker,
    ],
    'ascertainment': [
        BEASTAscertainmentChecker,
    ],
    'extra': [
        UnusualStateChecker,
        SingletonCharacterChecker,
    ],
}



import sys
import argparse

from termcolor import colored
from clldutils.clilib import PathType, ParserError

from nexus import NexusReader


def list_of_ranges(dstring):
    """
    Converts a comma-separated list of 1-based ranges into a list of 1-based indices.

    :param dstring: A string

    :return: `list` of zero-based indices
    """
    def _int(v):
        try:
            return int(v)
        except ValueError:
            raise argparse.ArgumentTypeError("%r is not an integer" % v)

    out = []
    for token in dstring.split(','):
        token = token.replace(':', '-')
        if '-' in token:
            start, stop = token.split("-")
            out.extend([x for x in range(_int(start), _int(stop) + 1)])
        else:
            out.append(_int(token))
    return sorted(out)


def path_or_stdin(string):
    return None if string == '-' else PathType(type='file')(string)


def add_nexus(parser, many=False):
    kw = {}
    if many:
        kw['nargs'] = '+'
    parser.add_argument(
        "filename",
        help='Path to Nexus file, or "-" to read from stdin',
        type=path_or_stdin,
        **kw)


def add_output(parser):
    parser.add_argument(
        "-o", "--output",
        default=None,
        help="output nexus file, if not specified, output will be printed to stdout. "
             "To prevent log messages messing up the output, set '--log-level=WARN'.")


def get_reader(args, many=False, required_blocks=None):
    res = []
    for f in (args.filename if many else [args.filename]):
        if f is None:
            res.append(NexusReader.from_string(sys.stdin.read()))
        else:
            res.append(NexusReader.from_file(f))
    if required_blocks:
        for nex in res:
            for block in required_blocks:
                if not getattr(nex, block, None):
                    raise ParserError(colored(
                        'Nexus file {0} has no {1} block'.format(nex.filename, block),
                        'red',
                        attrs=['bold'],
                    ))
    return res if many else res[0]


def write_output(writer, args):
    if args.output:
        writer.write_to_file(args.output)
        print('Output written to {0}'.format(args.output))
    else:
        print(writer.write())



"""
Performs a number of nexus counting/tallying methods.
"""
import textwrap
import _collections

from nexus.cli_util import add_nexus, get_reader
from nexus.tools import tally_by_taxon, tally_by_site, count_binary_set_size

TALLY_TYPES = _collections.OrderedDict([
    ('taxa', tally_by_taxon),
    ('site', tally_by_site),
    ('binary', count_binary_set_size),
])


def register(parser):
    add_nexus(parser)
    parser.add_argument(
        '-t', '--type',
        help='Typeof tally to compute',
        choices=list(TALLY_TYPES.keys()),
        default=list(TALLY_TYPES.keys())[0]
    )


def run(args):
    printer = print_binary if args.type == 'binary' else print_tally
    printer(TALLY_TYPES[args.type](get_reader(args)))


def print_tally(tally):
    wrapper = textwrap.TextWrapper(initial_indent=" ", subsequent_indent="\t", width=65)
    for tkey in sorted(tally):
        print(tkey)
        for skey in sorted(tally[tkey]):
            s = " ".join(sorted(["%s" % v for v in tally[tkey][skey]]))
            print(" - %s: " % skey, end="")
            for w in wrapper.wrap(s):
                print(w)
        print("\n")
    return


def print_binary(tally):
    for i in sorted(tally):
        print("%d\t%d" % (i, tally[i]))
    return



import pathlib
import functools

from nexus import NexusWriter, NexusReader


def get_nexus_reader(thing):
    if isinstance(thing, str):
        return NexusReader.from_string(thing)
    if isinstance(thing, pathlib.Path):
        return NexusReader.from_file(thing)
    if isinstance(thing, NexusWriter):
        return NexusReader.from_string(thing.write())
    assert isinstance(thing, NexusReader)
    return thing


def with_nexus_reader(func):
    @functools.wraps(func)
    def _(*args, **kwargs):
        return func(get_nexus_reader(args[0]), *args[1:], **kwargs)
    return _



import re

from nexus.exceptions import NexusFormatException
from nexus.handlers import GenericHandler
from nexus.handlers import QUOTED_PATTERN, END_PATTERN

TAXON_PLACEHOLDER = re.compile(r"""^\[.*\]\s+""")
TAXON_ANNOTATION = re.compile(r"""(.*)(\[.*\])$""")


class TaxaHandler(GenericHandler):
    """Handler for `taxa` blocks"""
    is_dimensions = re.compile(r"""dimensions\s*ntax\s*=\s*(\d+)""", re.IGNORECASE)
    is_taxlabel_block = re.compile(r"""\btaxlabels\b""", re.IGNORECASE)

    def __init__(self, **kw):
        super(TaxaHandler, self).__init__(**kw)
        self.taxa = []
        self.attributes = []
        self.annotations = {}

        in_taxlabel_block = False
        found_ntaxa = None
        for line in self.block:
            line = QUOTED_PATTERN.sub('\\1', line)
            if self.is_dimensions.match(line):
                found_ntaxa = int(self.is_dimensions.findall(line)[0])
                continue
            elif 'begin taxa' in line.lower():
                continue
            elif line == ';' or END_PATTERN.match(line.strip()):
                continue
            elif self.is_mesquite_attribute(line):
                self.attributes.append(line)
                continue
            elif in_taxlabel_block or self.is_taxlabel_block.match(line):
                in_taxlabel_block = True
                line = self.is_taxlabel_block.sub("", line)

            for taxon, annot in self._parse_taxa(line):
                self.taxa.append(taxon)
                if annot:
                    self.annotations[taxon] = annot

        if found_ntaxa and found_ntaxa != self.ntaxa:
            raise NexusFormatException(
                "Number of found taxa (%d) doesn't match dimensions declaration (%d)" % (
                    self.ntaxa, found_ntaxa))

    def __getitem__(self, index):
        return self.taxa[index]

    @property
    def ntaxa(self):
        return len(self.taxa)

    def _parse_taxa(self, line):
        taxa = [t.replace(";", "").strip() for t in line.split(" ")]
        for taxon in taxa:
            # remove initial comment
            taxon = TAXON_PLACEHOLDER.sub('', taxon)
            # get annotations
            annot = TAXON_ANNOTATION.match(taxon)
            if annot:
                taxon, annot = annot.groups()
            # remove quotes
            taxon = QUOTED_PATTERN.sub('\\1', taxon)

            if taxon and taxon not in self.taxa:
                yield (taxon, annot)

    def iter_lines(self):
        def wrap(s):
            return s if ' ' not in s else "'%s'" % s

        for att in self.attributes:
            yield "\t%s" % att
        yield '\tdimensions ntax=%d;' % self.ntaxa
        yield '\ttaxlabels'
        # taxa labels
        for idx, taxon in enumerate(self.taxa, 1):
            taxon = "%s%s" % (taxon, self.annotations.get(taxon, ''))
            yield "\t[%d] %s" % (idx, wrap(taxon))
        yield ';'



"""
Check nexus files for errors
"""
import warnings

from nexus.cli_util import add_nexus, get_reader
from nexus.checker import CHECKERS


def register(parser):
    add_nexus(parser)
    parser.add_argument(
        '-e', "--extra",
        help="add extra checks",
        action='store_true'
    )
    parser.add_argument(
        '-a', "--ascertainment",
        help="add ascertainment checks",
        action='store_true'
    )


def run(args):
    checkers = CHECKERS['base']
    if args.extra:
        checkers.extend(CHECKERS['extra'])
    if args.ascertainment:
        checkers.extend(CHECKERS['ascertainment'])

    with warnings.catch_warnings(record=True) as warned:
        warnings.simplefilter("always")
        nex = get_reader(args)

        if len(warned):
            print("Warnings encountered in reading nexus:")
            for w in warned:
                print("\t%s" % w)

    for checker in checkers:
        checker(nex).status()



import random
import functools

from ..handlers.tree import Tree
from .util import with_nexus_reader


def replace_trees(func):
    """
    Decorator for generators yielding a mutated set of trees for a nexus object.

    :param func: A generator function that yields the new `Tree` objects.
    """
    @functools.wraps(func)
    def _(nexus_obj, *args, **kwargs):
        new = []
        for tree in func(nexus_obj, *args, **kwargs):
            assert isinstance(tree, Tree), 'tree manipulators must yield nexus.handlers.tree.Tree s'
            new.append(tree)
        nexus_obj.trees.trees = new
        return nexus_obj
    return _


@with_nexus_reader
@replace_trees
def visit_trees(nexus_obj, visitor, log=None):
    """
    Manipulate all trees in a `NexusReader` by running a callable with the following signature:

    .. code-block:: python

        def visitor(tree: newick.Node) -> typing.Union[newick.Node, str, None]: pass

    If the visitor returns `None`, the tree is deleted, otherwise replaced with the returned
    newick representation.
    """
    for tree in nexus_obj.trees:
        res = visitor(tree.newick_tree)
        if res:
            yield Tree.from_newick(res, name=tree.name, rooted=tree.rooted)


@with_nexus_reader
@replace_trees
def visit_tree_nodes(nexus_obj, visitor, log=None):
    """
    Manipulate all trees in a `NexusReader` by running a callable on each node of each tree.

    :param visitor: callable suitable for passing into `newick.Node.visit`.
    """
    for tree in nexus_obj.trees:
        ntree = tree.newick_tree
        ntree.visit(visitor)
        yield Tree.from_newick(ntree, name=tree.name, rooted=tree.rooted)


@with_nexus_reader
@replace_trees
def delete_trees(nexus_obj, delitems, log=None):
    """
    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A NexusReader instance with the given trees removed.
    """
    if log:
        log.info('Deleting: %d trees' % len(delitems))

    for index, tree in enumerate(nexus_obj.trees, 1):
        if index in delitems:
            if log:
                log.info('Deleting tree %d' % index)
        else:
            yield tree


@with_nexus_reader
@replace_trees
def sample_trees(nexus_obj, num_trees=None, every_nth=None, log=None):
    """
    Returns a specified number (`num_trees`) of random trees from the nexus.

    :param num_trees: The number of trees to resample
    :type num_trees: Integer

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A NexusReader instance.

    :raises ValueError: if num_trees is larger than population
    """
    assert (num_trees or every_nth) and not (num_trees and every_nth), \
        "One of num_trees and every_nth must be selected"
    if num_trees:
        if num_trees > nexus_obj.trees.ntrees:
            raise ValueError("Treefile only has %d trees in it." % nexus_obj.trees.ntrees)
        if num_trees == nexus_obj.trees.ntrees:  # pragma: no cover
            return nexus_obj  # um. ok.
        trees = random.sample(nexus_obj.trees.trees, num_trees)
    else:
        trees = [tree for index, tree in enumerate(nexus_obj.trees, 1) if index % every_nth == 0]

    if log:
        log.info("%d trees read. Sampling %d" % (nexus_obj.trees.ntrees, len(trees)))
    yield from trees


@with_nexus_reader
@replace_trees
def strip_comments_in_trees(nexus_obj, log=None):
    """
    Removes comments from the trees in a nexus

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A NexusReader instance with the comments removed.
    """
    for tree in nexus_obj.trees:
        yield Tree(nexus_obj.trees.remove_comments(tree))

    if log:
        log.info("Removed comments")



"""
Describes the given character.
"""
import textwrap
import collections

from nexus.cli_util import add_nexus, get_reader

wrapper = textwrap.TextWrapper(initial_indent="  ", subsequent_indent="  ")


def register(parser):
    parser.add_argument('site_index', help="Character index")
    add_nexus(parser)


def run(args):
    print_character_stats(get_reader(args, required_blocks=['data']), args.site_index)


def print_character_stats(nexus_obj, character_index):
    """
    Prints the character/site statistics for a given `nexus_obj` and
    character index

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param character_index: The character index of the character to summarise
    :type character_index: Int or String

    :raises IndexError: if character_index is not in nexus data block
    """
    index = None
    if character_index in nexus_obj.data.characters:
        index = character_index  # string index
    else:
        try:
            character_index = int(character_index)
        except ValueError:  # pragma: no cover
            pass

        if character_index in nexus_obj.data.characters:
            index = character_index

    if index is None:
        raise IndexError("Character '%s' is not in the nexus" % character_index)

    states = collections.defaultdict(list)
    for taxon, state in nexus_obj.data.characters[index].items():
        states[state].append(taxon)

    for state in sorted(states):
        print('State: %s (%d / %d = %0.2f)' % (
            state,
            len(states[state]), nexus_obj.data.ntaxa,
            (len(states[state]) / nexus_obj.data.ntaxa * 100)
        ))
        print("\n".join(wrapper.wrap(", ".join(sorted(states[state])))))
        print("\n")
    return



"""
Displays some simple nexus info
"""
import collections

from clldutils.clilib import Table, add_format

from nexus.cli_util import add_nexus, get_reader


def register(parser):
    add_nexus(parser)
    add_format(parser, default='simple')


def run(args):
    nexus_obj = get_reader(args, required_blocks=['data'])
    print(nexus_obj.filename)

    with Table(args, 'Taxon', 'Characters') as t:
        for taxon in sorted(nexus_obj.data.matrix):
            tally = collections.Counter()
            for site in nexus_obj.data.matrix[taxon]:
                tally.update([site])

            t.append([taxon, ", ".join(['%s x %s' % (k, tally[k]) for k in sorted(tally)])])



import pathlib


class FileWriterMixin(object):
    def write_to_file(self, filename, encoding='utf8', **kw):
        """
        Writes the nexus to a file.

        :return: `pathlib.Path` instance of the written file.
        """
        res = pathlib.Path(filename)
        res.write_text(self.write(**kw), encoding=encoding)
        return res



from nexus.writer import NexusWriter
from .util import with_nexus_reader


def _recode_to_binary(char, keep_zero=False):
    """
    Recodes a dictionary to binary data.

    :param char: A dictionary of taxa to state values
    :type char: dict

    :param keep_zero: A boolean flag denoting whether to
        treat '0' as a missing state or not. The default
        (False) is to ignore '0' as a trait absence.

        Setting this to True will treat '0' as a unique
        state.
    :type keep_zero: Boolean

    :return: A dictionary of taxa to recoded values.
    :raises ValueError: if any of the states in the
        `char` dictionary is not a string (i.e.
        integer or None values)

    >>> states, recode = _recode_to_binary({'Maori': '1', 'Dutch': '2', 'Latin': '1'})
    >>> states
    ['1', '2']
    >>> recode['Maori']
    '10'
    >>> recode['Dutch']
    '01'
    >>> recode['Latin']
    '10'
    """
    newdata = {}

    # unwanted states
    unwanted_states = ['-', '?']
    if not keep_zero:
        unwanted_states.append('0')

    if not all(isinstance(v, str) for v in char.values()):
        raise ValueError('Data must be strings: %r' % char.values())

    # preprocess taxa states and get unique states
    states = set()
    for taxon, value in char.items():
        char[taxon] = [v for v in value.replace(" ", ",").split(",") if v not in unwanted_states]
        states.update(char[taxon])

    states = tuple(sorted(states))
    num_states = len(states)
    for taxon, values in char.items():
        newdata[taxon] = ['0' for _ in range(num_states)]
        for value in values:
            if value not in unwanted_states:  # ignore missing values
                newdata[taxon][states.index(value)] = '1'
        newdata[taxon] = "".join(newdata[taxon])
        assert len(newdata[taxon]) == num_states

    return (states, newdata)


@with_nexus_reader
def binarise(nexus_obj, keep_zero=False):
    """
    Returns a binary variant of the given `nexus_obj`.
    If `one_nexus_per_block` then we return a list of NexusWriter instances.

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param keep_zero: A boolean flag denoting whether to
        treat '0' as a missing state or not. The default
        (False) is to ignore '0' as a trait absence.

        Setting this to True will treat '0' as a unique
        state.
    :type keep_zero: Boolean

    :return: A NexusWriter instance or a list of NexusWriter instances.
    :raises AssertionError: if nexus_obj is not a nexus
    :raises NexusFormatException: if nexus_obj does not have a `data` block
    """
    n = NexusWriter()

    for i in sorted(nexus_obj.data.charlabels):
        label = nexus_obj.data.charlabels[i]  # character label
        char = nexus_obj.data.characters[label]  # character dict
        states, recoding = _recode_to_binary(char, keep_zero)  # recode

        new_char_length = len(recoding[list(recoding.keys())[0]])

        # loop over recoded data
        for j in range(new_char_length):
            for taxon, value in recoding.items():
                # make new label
                new_label = "%s_%s" % (str(label), states[j])
                # add to nexus
                n.add(taxon, new_label, value[j])
    return n




class NexusFormatException(Exception):
    """Generic Exception for Nexus Format Errors"""
    pass


class TranslateTableException(NexusFormatException):
    """Exception for Translate table Errors"""
    pass



import collections

from nexus.tools.sites import new_nexus_without_sites
from .util import with_nexus_reader


@with_nexus_reader
def check_zeros(nexus_obj, absences=None, missing=None):
    """
    Checks for sites in the nexus that are coded as all empty.

    Returns a list of sites that are completely empty. Note that
    this is zero-indexed (i.e. the first site is site 0 not 1)
    to enable indexing of nexus.data.matrix/.character lists

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param absences: A list of values to be marked as absent.
        Default = ["0"]
    :type char: list

    :param missing: A list of values to be marked as missing.
        Default = ["-", "?"]
    :type char: list

    :return: A list of site indexes
    :raises ValueError: if any of the states in the
        `char` dictionary is not a string (i.e.
        integer or None values)
    """
    absences = absences if absences else ['0']
    missing = missing if missing else ['-', '?']

    bad = []
    for site_idx in range(0, nexus_obj.data.nchar):
        states = collections.Counter(
            [nexus_obj.data.matrix[t][site_idx] for t in nexus_obj.data.matrix])
        zeros = sum([states[k] for k in states if k in absences or k in missing])
        total = sum(states.values())
        if zeros == total:
            bad.append(site_idx)
    return bad


@with_nexus_reader
def remove_zeros(nexus_obj, absences=None, missing=None):
    """
    Removes sites in the nexus that are coded as all empty.

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param absences: A list of values to be marked as absent.
        Defacult = ["0"]
    :type char: list

    :param missing: A list of values to be marked as missing.
        Default = ["-", "?"]
    :type char: list

    :return: a new nexus
    """
    zeros = check_zeros(nexus_obj, absences=absences, missing=missing)
    return new_nexus_without_sites(nexus_obj, zeros)._convert_to_reader()



"""
Performs some functions on trees
"""
from clldutils.clilib import add_random_seed
from nexus.cli_util import add_nexus, get_reader, add_output, write_output, list_of_ranges
from nexus.tools import delete_trees, sample_trees, strip_comments_in_trees


def register(parser):
    add_output(parser)
    add_nexus(parser)
    parser.add_argument(
        "-d", "--deltree",
        default=[],
        type=list_of_ranges,
        help="Remove the trees specified as comma-separated ranges of 1-based indices, "
             "e.g. '1', '1-5', '1,20-30'")
    parser.add_argument(
        "-r", "--resample",
        type=int,
        default=0,
        help="Resample the trees every Nth tree")
    parser.add_argument(
        "-n", "--random",
        type=int,
        default=0,
        help="Randomly sample N trees from the treefile")
    add_random_seed(parser)
    parser.add_argument(
        "-c", "--removecomments",
        action="store_true",
        default=False,
        help="Remove comments from the trees")
    parser.add_argument(
        "-t", "--detranslate",
        action="store_true",
        default=False,
        help="Remove taxa translation block from the trees")


def run(args):
    nexus = get_reader(args, required_blocks=['trees'])
    args.log.info("{0} trees found with {1} translated taxa".format(
        nexus.trees.ntrees, len(nexus.trees.translators)))

    if args.deltree:
        nexus = delete_trees(nexus, args.deltree, log=args.log)

    if args.resample:
        nexus = sample_trees(nexus, every_nth=args.resample, log=args.log)

    if args.random:
        nexus = sample_trees(nexus, num_trees=args.random, log=args.log)

    if args.removecomments:
        nexus = strip_comments_in_trees(nexus, log=args.log)

    if args.detranslate:
        nexus.trees.detranslate()

    write_output(nexus, args)



"""
Convert a nexus file to a different format
"""
import textwrap

from nexus.cli_util import add_nexus, get_reader, add_output, write_output
from nexus.util import FileWriterMixin


def register(parser):
    add_nexus(parser)
    add_output(parser)
    parser.add_argument('--format', choices=['fasta'], default='fasta', help='Output format')


def run(args):
    write_output(Converter(get_reader(args)), args)


class Converter(FileWriterMixin):
    def __init__(self, nex):
        self.nex = nex

    def write(self, format='fasta', **kw):
        res = []
        if format == 'fasta':
            for taxon in sorted(self.nex.data.matrix):
                res.append('>%s' % taxon)
                for line in textwrap.wrap("".join(self.nex.data.matrix[taxon]), 70):
                    res.append(line)
        else:  # pragma: no cover
            raise NotImplementedError(format)

        return '\n'.join(res)



"""
Performs a number of nexus manipulation methods.
"""
import collections

from nexus.cli_util import add_nexus, get_reader, add_output, write_output, list_of_ranges
from nexus.tools import count_site_values
from nexus.tools import check_zeros
from nexus.tools import iter_constant_sites
from nexus.tools import iter_unique_sites
from nexus.tools import new_nexus_without_sites


def register(parser):
    add_output(parser)
    add_nexus(parser)
    parser.add_argument(
        "-n", "--number",
        action="store_true",
        default=False,
        help="Count the number of characters")
    parser.add_argument(
        "-c", "--constant",
        action="store_true",
        default=False,
        help="Remove the constant characters")
    parser.add_argument(
        "-u", "--unique",
        action="store_true",
        default=False,
        help="Remove the unique characters")
    parser.add_argument(
        "-x", "--remove",
        type=list_of_ranges,
        help="Remove characters specified as comma-separated ranges of 1-based indices, "
             "e.g. '1', '1-5', '1,20-30'")
    parser.add_argument(
        "-z", "--zeros",
        action="store_true",
        default=False,
        help="Remove the zero characters")
    parser.add_argument(
        "-s", "--stats",
        action="store_true",
        default=False,
        help="Print character-by-character stats")


def run(args):
    nexus = get_reader(args)

    if args.number:
        print_site_values(nexus)
        return 0

    if args.stats:
        print_character_stats(nexus)
        return 0

    const, unique, zeros, remove = [], [], [], []
    if args.constant:
        const = list(iter_constant_sites(nexus))
        if const:
            args.log.info("Constant Sites: %s" % ",".join([str(i + 1) for i in const]))
    if args.unique:
        unique = list(iter_unique_sites(nexus))
        if unique:
            args.log.info("Unique Sites: %s" % ",".join([str(i + 1) for i in unique]))
    if args.zeros:
        zeros = check_zeros(nexus)
        if zeros:
            args.log.info("Zero Sites: %s" % ",".join([str(i + 1) for i in zeros]))
    if args.remove:
        args.log.info("Remove: %s" % ",".join([str(i) for i in args.remove]))
        remove = [i - 1 for i in args.remove]  # translate to zero-based indices.

    write_output(new_nexus_without_sites(nexus, set(const + unique + zeros + remove)), args)


def print_site_values(nexus_obj, characters=None):
    """
    Prints out counts of the number of sites with state in `characters` in a
    nexus.

    (Wrapper around `count_site_values`)

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader
    """
    characters = characters if characters is not None else ['-', '?']

    count = count_site_values(nexus_obj, characters)
    print("Number of %s in %s" % (",".join(characters), nexus_obj.filename))
    for taxon in sorted(count):
        prop = (count[taxon] / nexus_obj.data.nchar) * 100
        print(
            "%s: %d/%d (%0.2f%%)" %
            (taxon.ljust(20), count[taxon], nexus_obj.data.nchar, prop)
        )
    print('-' * 76)
    total_count = sum([x for x in count.values()])
    total_data = nexus_obj.data.nchar * nexus_obj.data.ntaxa
    prop = (total_count / total_data) * 100
    print('TOTAL: %d/%d (%0.2f%%)' %
          (total_count, total_data, prop)
          )


def print_character_stats(nexus_obj):
    """
    Prints the number of states and members for each site in `nexus_obj`

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :return: A list of the state distribution
    """
    state_distrib = []
    for i in range(0, nexus_obj.data.nchar):
        tally = collections.Counter()
        for taxa, characters in nexus_obj.data:
            tally.update([characters[i]])

        print("%5d" % i, end="")
        for state in tally:
            print("%sx%d" % (state, tally[state]), end="")
            state_distrib.append(tally[state])
        print("\n")
    return state_distrib



"""
Converts a multistate nexus file to binary present/absent form.
"""
from nexus.tools.binarise import binarise
from nexus.cli_util import add_nexus, get_reader, add_output, write_output


def register(parser):
    add_output(parser)
    add_nexus(parser)


def run(args):
    write_output(binarise(get_reader(args)), args)



import re

COMMENT_PATTERN = re.compile(r"""(\[.*?])""")
QUOTED_PATTERN = re.compile(r"""^["'](.*)["']$""")
WHITESPACE_PATTERN = re.compile(r"""\s+""")
MESQUITE_TITLE_PATTERN = re.compile(r"""^TITLE\s+(.*);$""", re.IGNORECASE)
MESQUITE_LINK_PATTERN = re.compile(
    r"""^LINK\s+(.*?)\s+=\s+(.*);$""", re.IGNORECASE
)
BEGIN_PATTERN = re.compile(r"""begin (\w+)(\s*|\[.*]);""", re.IGNORECASE)
END_PATTERN = re.compile(r"""end\s*;""", re.IGNORECASE)


class GenericHandler(object):
    """
    Handlers are objects to store specialised blocks found in nexus files.

    Nexus Block->Handler mapping is initialised in Nexus.handlers

    Handlers have (at least) the following attributes:

        1. __init__(self, **kw) - the function for parsing the block
        2. iter_lines(self) - a function for returning the block to a text
            representation (used to regenerate a nexus file).
        3. block - a list of raw strings in this block
    """
    def __init__(self, name=None, data=None):
        """Initialise datastore in <block> under <keyname>"""
        self.name = name
        self.block = data or []
        self.comments = []

        # save comments
        for line in self.block:
            if line.strip().startswith("[") and line.strip().endswith("]"):
                self.comments.append(line)

    def iter_lines(self):
        for i, line in enumerate(self.block):
            if (i == 0 and BEGIN_PATTERN.search(line)) or \
                    (i == len(self.block) - 1 and END_PATTERN.search(line)):
                continue
            yield line

    def write(self):
        """
        Generates a string containing a nexus block.
        """
        return "".join(
            ['begin {0};\n'.format(self.name)] +  # noqa: W504
            [line + '\n' for line in self.iter_lines()] +  # noqa: W504
            ['end;\n']
        )

    @staticmethod
    def remove_comments(line):
        """
        Removes comments from lines

        >>> GenericHandler.remove_comments("Hello [world]")
        'Hello '
        >>> GenericHandler.remove_comments("He[bite]ll[me]o")
        'Hello'

        :param line: string
        :type line: string

        :return: Returns a cleaned string.
        """
        return COMMENT_PATTERN.sub('', line)

    @staticmethod
    def is_mesquite_attribute(line):
        """
        Returns True if the line is a mesquite attribute
        """
        return bool(MESQUITE_TITLE_PATTERN.match(line)) or bool(MESQUITE_LINK_PATTERN.match(line))



"""
Converts an interleaved nexus to a simple nexus.

See http://mrbayes.sourceforge.net/Help/matrix.html
"""
from nexus.cli_util import add_nexus, get_reader, add_output, write_output


def register(parser):
    add_output(parser)
    add_nexus(parser)


def run(args):
    write_output(get_reader(args), args)



import re
import warnings
import collections

from nexus.handlers import GenericHandler
from nexus.handlers import QUOTED_PATTERN, WHITESPACE_PATTERN, BEGIN_PATTERN, END_PATTERN

NTAX_PATTERN = re.compile(r"""ntax=(\d+)""", re.IGNORECASE)
NCHAR_PATTERN = re.compile(r"""nchar=(\d+)""", re.IGNORECASE)


def iter_block(lines):
    seen_matrix = False
    for line in lines:
        lline = line.lower().strip()
        if END_PATTERN.match(lline):
            continue
        elif lline.startswith('format'):
            continue
        elif BEGIN_PATTERN.match(line):
            continue
        elif lline.startswith('matrix'):
            seen_matrix = True
            continue
        yield line, lline, seen_matrix


class DataHandler(GenericHandler):
    """Handler for data matrices"""

    _character_block_pattern = re.compile(
        r"""charstatelabels(.*?);""",
        re.IGNORECASE | re.DOTALL
    )
    _format_line_pattern = re.compile(
        r"""format\b(.*?);""",
        re.IGNORECASE | re.DOTALL | re.MULTILINE
    )
    _label_pattern = re.compile(
        r"""(.+?)=(.+?);"""
    )

    def __init__(self, **kw):
        super(DataHandler, self).__init__(**kw)
        self.charlabels = {}
        self.attributes = []
        self.format = {}
        self.gaps = None
        self.missing = None
        self.matrix = collections.defaultdict(list)
        self._sitecache = {}  # cache for site patterns to parsed sites
        self._characters = None  # cache for characters list
        self._symbols = None  # cache for symbols list

        self.format = self.parse_format_line("\n".join(self.block))
        self.block = self._parse_charstate_block(self.block)
        _dim_taxa, _dim_chars = None, None

        read_data = False
        for line, lline, in_matrix in iter_block(self.block):
            # Dimensions line
            if lline.startswith('dimensions '):
                try:  # try for ntaxa
                    _dim_taxa = int(NTAX_PATTERN.findall(line)[0])
                except IndexError:  # pragma: no cover
                    pass
                try:  # and nchar
                    _dim_chars = int(NCHAR_PATTERN.findall(line)[0])
                except IndexError:  # pragma: no cover
                    pass
            elif self.is_mesquite_attribute(line):
                self.attributes.append(line)

            elif in_matrix:
                line = self.remove_comments(line)
                try:  # NORMALISE WHITESPACE
                    taxon, sites = WHITESPACE_PATTERN.split(line, 1)
                    read_data = True
                except ValueError:
                    continue

                taxon = QUOTED_PATTERN.sub('\\1', taxon.strip())
                self.add_taxon(taxon, self._parse_sites(sites.strip()))
            elif self._label_pattern.findall(line):
                for label in self._label_pattern.findall(line):
                    for site in self.parse_range(label[1]):
                        # note we subtract 1 from the site index in the nexus
                        # file as this is zero indexed in DataHandler.matrix.
                        self.charlabels[site - 1] = label[0]

        if not read_data:
            # Let's try to read a "wrapped" matrix:
            taxon, sites = None, []
            for line, lline, in_matrix in iter_block(self.block):
                if (not in_matrix) or (not lline):
                    continue  # pragma: no cover
                if not taxon:
                    assert not WHITESPACE_PATTERN.search(line.strip())
                    taxon = QUOTED_PATTERN.sub('\\1', line.strip())
                else:
                    sites.extend(self._parse_sites(line.strip()))
                    if len(sites) == _dim_chars:
                        self.add_taxon(taxon, sites)
                        taxon, sites = None, []

        # Warn if format string (ntaxa or nchar) does not give the right answer
        if _dim_taxa is not None and self.ntaxa != _dim_taxa:
            warnings.warn("Expected %d taxa, got %d" % (self.ntaxa, _dim_taxa))

        if _dim_chars is not None and self.nchar != _dim_chars:
            warnings.warn("Expected %d characters, got %d" % (self.nchar, _dim_chars))

    def __getitem__(self, index):
        return self.taxa[index], self.matrix.get(self.taxa[index])

    def __repr__(self):
        return '<NexusDataBlock: {0.nchar} characters from {0.ntaxa} taxa>'.format(self)

    @property
    def ntaxa(self):
        """Number of Taxa"""
        return len(self.matrix)

    @property
    def nchar(self):
        """Number of Characters"""
        if self.taxa and self.matrix:
            return len(self.matrix[self.taxa[0]])
        elif self.charlabels:
            return len(self.charlabels)
        return 0  # pragma: no cover

    @property
    def taxa(self):
        """Taxa list"""
        return list(self.matrix.keys())

    @property
    def symbols(self):
        """Distinct symbols in matrix"""
        if not self._symbols:
            self._symbols = set()
            [self._symbols.update(vals) for vals in self.matrix.values()]
        return self._symbols

    @property
    def characters(self):
        if not self._characters:
            self._characters = collections.defaultdict(dict)
            for taxon in self.taxa:
                for index, _ in enumerate(self.matrix[taxon]):
                    label = self.charlabels.get(index, index)
                    self._characters[label][taxon] = self.matrix[taxon][index]
        return self._characters

    def is_missing_or_gap(self, state):
        return state in ('-', '?')

    def parse_format_line(self, data):
        """
        Parses a format line, and returns a dictionary of tokens

        :param data: string
        :type data: string

        :return: Returns a dictionary of tokens in the format line.
        """
        out = {}

        try:
            line = self._format_line_pattern.findall(data)[0]
        except IndexError:
            return None

        line = line.replace(" =", "=").replace("= ", "=")  # standardise
        for chunk in WHITESPACE_PATTERN.split(line):
            try:
                key, value = chunk.split("=", maxsplit=1)
                value = QUOTED_PATTERN.sub('\\1', value)
            except ValueError:
                key, value = chunk, True
            if key:
                out[key.lower()] = value
        return out

    def _parse_sites(self, sites):
        """
        Parses a string of sites and returns a list of site values

        >>> DataHandler()._parse_sites('123')
        ['1', '2', '3']
        >>> DataHandler()._parse_sites('1(12)')
        ['1', '12']
        >>> DataHandler()._parse_sites('123(4,5)56')
        ['1', '2', '3', '4,5', '5', '6']
        >>> DataHandler()._parse_sites('123(4 5)56')
        ['1', '2', '3', '4 5', '5', '6']
        >>> DataHandler()._parse_sites("ACGTU?")
        ['A', 'C', 'G', 'T', 'U', '?']

        :param sites: string
        :type sites: string

        :return: Returns a list of site values
        :raises NexusFormatException: If data matrix contains incomplete
            multistate values
        """
        remove_values = [' ', ';']
        parsed = [s for s in sites if s not in remove_values]
        if sites not in self._sitecache:
            # Slow parser for multistate
            if '(' in parsed:
                todo, parsed = parsed, []  # switch places.
                while todo:
                    site = todo.pop(0)
                    if site == ',':
                        continue
                    elif site == '(':
                        # read-ahead
                        site = ''  # discard open bracket
                        multistate = True
                        while multistate:
                            nextchar = todo.pop(0)
                            if nextchar == ')':
                                multistate = False
                            else:
                                site += nextchar
                    parsed.append(site)
                # end slow parser
        self._sitecache[sites] = parsed
        return self._sitecache[sites]

    def add_taxon(self, taxon, site_values=None):
        """
        Adds a `taxon` to the matrix with values from `site_values`

        :param taxon: taxa name
        :type data: string

        :param site_values: site values
        :type data: list

        :return: None
        """
        self.matrix[taxon].extend(site_values)

    def del_taxon(self, taxon):
        """
        Removes `taxon` from the data

        :param taxon: taxa name
        :type data: string

        :return: None
        """
        del(self.matrix[taxon])

    def _parse_charstate_block(self, data):
        """
        Extracts the character state block and returns the remaining matrix
        """
        char_number_pattern = re.compile(r"""(\d+)\s+(.*)""")
        char_index = 0

        new_data = "\n".join(data)
        charblock = self._character_block_pattern.findall(new_data)
        new_data = self._character_block_pattern.sub('', new_data)
        if len(charblock) == 1:
            charblock = charblock[0]
            for char in charblock.split(","):
                char = char.strip()
                if not len(char):
                    warnings.warn(
                        "Encountered an empty character label at position %d" % char_index)
                    continue
                char = char_number_pattern.sub('\\2', char)
                self.charlabels[char_index] = char
                char_index += 1
        return new_data.split("\n")

    def parse_range(self, siterange):
        if ',' in siterange:
            return [int(c) for c in siterange.split(',')]
        elif '-' in siterange:
            start, stop = siterange.split('-')
            start, stop = int(start), int(stop)
            return list(range(start, stop + 1))
        else:
            return [int(siterange)]

    def iter_lines(self):
        """
        Generates a string containing a nexus data block.

        :return: String
        """
        def _make_format_line(self):
            """
            Generates a format string.

            :return: String
            """
            fstring = ['\tformat']
            for key in sorted(self.format):
                value = self.format[key]
                if key == 'datatype':
                    # Datatype must come first!
                    fstring.insert(1, "%s=%s" % (key, value))
                elif key in ('interleave', ):
                    # IGNORE the interleaving -- not implemented
                    continue
                else:
                    if key == 'symbols':
                        value = '"%s"' % "".join(sorted([
                            s for s in self.symbols if not self.is_missing_or_gap(s)
                        ]))
                    fstring.append("%s=%s" % (key, value))
            return " ".join(fstring) + ";"

        for att in self.attributes:
            yield "\t%s" % att
        yield '\tdimensions ntax=%d nchar=%d;' % (self.ntaxa, self.nchar)
        yield _make_format_line(self)
        # handle char block
        if self.charlabels:
            yield '\tcharstatelabels'
            max_id = max(self.charlabels)
            for char_id in sorted(self.charlabels):
                yield '\t\t%d %s%s' % (  # zero-indexing
                    char_id + 1, self.charlabels[char_id], ',' if char_id < max_id else '')
            yield '\t;'
        yield "matrix"
        max_taxon_len = max([len(_) for _ in self.matrix])
        for taxon in sorted(self.matrix):
            yield "%s %s" % (taxon.ljust(max_taxon_len), ''.join(self.matrix[taxon]))
        yield " ;"


class CharacterHandler(DataHandler):
    pass



from setuptools import setup, find_packages


setup(
    name="python-nexus",
    version="2.8.0",
    description="A nexus (phylogenetics) file reader (.nex, .trees)",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="phylogenetics nexus newick paup splitstree",
    author="Simon Greenhill and Robert Forkel",
    author_email="simon@simon.net.nz",
    url="https://github.com/dlce-eva/python-nexus",
    license="BSD-2-Clause",
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    test_suite="tests",
    entry_points={
        'console_scripts': [
            'nexus=nexus.__main__:main',
        ],
    },
    platforms='any',
    python_requires='>=3.5',
    install_requires=[
        'newick>=1.3.1',
        'clldutils>=3.10.0',
        'termcolor',
    ],
    extras_require={
        'dev': ['flake8', 'wheel', 'twine'],
        'test': [
            'pytest>=5',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
        ],
    },
)



"""
python-nexus - Generic nexus (.nex, .trees) reader for python
"""
from nexus.reader import NexusReader
from nexus.writer import NexusWriter
from nexus import handlers
from nexus.exceptions import NexusFormatException
from nexus import tools

__version__ = "2.8.0"
__all__ = ["NexusReader", "NexusWriter", "NexusFormatException", "handlers", "tools"]



import random

from nexus.writer import NexusWriter
from .util import with_nexus_reader


@with_nexus_reader
def shufflenexus(nexus_obj, resample=False):
    """
    Shuffles the characters between each taxon to create a new nexus

    :param nexus_obj: A `NexusReader` instance
    :type nexus_obj: NexusReader

    :param resample: The number of characters to resample. If set to False,
        then the number of characters will equal the number of characters
        in the original data file.
    :type resample: Integer

    :return: A shuffled NexusReader instance
    :raises ValueError: if resample is not False or a positive Integer
    """
    if resample is False:
        resample = nexus_obj.data.nchar

    try:
        resample = int(resample)
    except ValueError:
        raise ValueError('resample must be a positive integer or False!')

    if resample < 1:
        raise ValueError('resample must be a positive integer or False!')

    newnexus = NexusWriter()
    newnexus.add_comment("Randomised Nexus generated from %s" % nexus_obj.filename)

    for i in range(resample):
        # pick existing character
        character = random.randrange(0, nexus_obj.data.nchar)
        chars = nexus_obj.data.characters[character]
        site_values = [chars[taxon] for taxon in nexus_obj.data.taxa]
        random.shuffle(site_values)
        for taxon in nexus_obj.data.taxa:
            newnexus.add(taxon, i, site_values.pop(0))
    return newnexus
